/* Simple Xlib application creating a window 
   --- the hard way, using only system calls
   instead of libX11. Compare to "x11-lib.c". */

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/un.h>

/* I/O helpers */

void writen(int s, void *p, int n) {
  if (write(s, p, n) != n) {
    perror("write failed");
    abort();
  }
}

void write8(int s, unsigned char c) {
  writen(s, &c, 1);
}

void write16(int s, unsigned short c) {
  writen(s, &c, 2);
}

void write32(int s, unsigned int c) {
  writen(s, &c, 4);
}

void readn(int s, void *p, int amt)
{
  if (read(s, p, amt) != amt) {
    perror("read failed");
    abort();
  }
}

int read8(int s)
{
  unsigned char b[1];
  readn(s, b, 1);
  return b[0];
}

int read16(int s)
{
  unsigned short b[1];
  readn(s, b, 2);
  return b[0];
}

/* Structures that reflect the X11 protocol: */

typedef struct x11_connect_info {
  int release_num;
  int resource_id_base;
  int resource_id_mask;
  int motion_buffer_size;
  unsigned short vendor_length;
  unsigned short max_req_length;
  char num_roots;
  char num_formats;
  char image_lsb_first;
  char bitmap_lsb_first;
  unsigned char bitmap_format_unit;
  unsigned char bitmap_format_pad;
  unsigned char min_keycode;
  unsigned char max_keycode;
  char unused[4];
} x11_connect_info;

typedef struct x11_root {
  unsigned int id;
  unsigned int colormap_id;
  unsigned int white_pixel;
  unsigned int black_pixel;
  unsigned int event_mask_at_open;
  unsigned short width;
  unsigned short height;
  unsigned short width_mm;
  unsigned short height_mm;
  unsigned short min_installed_maps;
  unsigned short max_installed_maps;
  unsigned int visual;
  unsigned char backing_stored;
  unsigned char save_unders;
  unsigned char root_depth;
  unsigned char num_depths;
} x11_root;

typedef struct x11_visual {
  unsigned int visual;
  unsigned char class;
  unsigned char bits_per_rgb;
  unsigned short colormap_entries;
  unsigned int red_mask;
  unsigned int green_mask;
  unsigned int blue_mask;
  char unuused[4];
} x11_visual;

typedef struct x11_create_window {
  unsigned char req; // = 1
  unsigned char depth;
  unsigned short length;
  unsigned int wid;
  unsigned int parent;
  short x;
  short y;
  unsigned short width;
  unsigned short height;
  unsigned short border_width;
  unsigned short class;
  unsigned int visual;
  unsigned int mask;
} x11_create_window;

typedef struct x11_map_window {
  unsigned char req; // = 8
  unsigned char pad;
  unsigned short length;
  unsigned int id;
} x11_map_window;

/* Make a connection and send raw protocol bytes: */

int main()
{
  int s, r, total_len, i, j, k;
  int depth, num_visuals;
  unsigned int id = 0x4000001;
  struct sockaddr_un addr;
  x11_connect_info ci;
  x11_root rt;
  x11_visual vis;
  x11_create_window cw;
  x11_map_window mw;
  char *v;
  
  s = socket(AF_UNIX, SOCK_STREAM, 0);
  memset(&addr, 0, sizeof(addr));
  addr.sun_family = AF_UNIX;
  v = getenv("DISPLAY");
  if ((v[0] == ':')
      && (v[1] == '0'))
    strcpy(addr.sun_path, "/tmp/.X11-unix/X0");
  else
    strcpy(addr.sun_path, v);
  
  r = connect(s, (struct sockaddr *)&addr, sizeof(addr));
  if (r == -1) {
    perror("connect failed");
    abort();
  }

  /* initial connection: */
  write8(s, 0x6c); /* little-endian */
  write8(s, 0);
  write16(s, 11);
  write16(s, 0);
  write16(s, 0); /* authname len */
  write16(s, 0); /* authdata len */
  write16(s, 0);

  if (read8(s) != 1) {
    perror("X11 connect failed");
    abort();
  }
  read8(s); /* reason len */
  read16(s); /* major version */
  read16(s); /* minor version */
  total_len = read16(s);

  /* Receive connection information from the server: */

  readn(s, &ci, sizeof(ci));

  v = malloc(ci.vendor_length);
  readn(s, v, ci.vendor_length);

  for (i = 0; i < ci.num_formats; i++) {
    read8(s); /* depth */
    read8(s); /* bits-per-pixel */
    read8(s); /* pad */
    read8(s);read8(s);read8(s);read8(s);read8(s);
  }

  for (i = 0; i < ci.num_roots; i++) {
    readn(s, &rt, sizeof(rt));    
    for (j = 0; j < rt.num_depths; j++) {
      depth = read16(s);
      num_visuals = read16(s);
      read16(s); read16(s);
      for (k = 0; k < num_visuals; k++) {
        readn(s, &vis, sizeof(vis));
      }
    }
  }

  /* Create a window */

  id = ci.resource_id_base + 1;

  cw.req = 1; /* 1 means "new window */
  cw.depth = 0;
  cw.length = sizeof(cw) >> 2;
  cw.wid = id;
  cw.parent = rt.id;
  cw.x = 10;
  cw.y = 10;
  cw.width = 100;
  cw.height = 100;
  cw.border_width = 1;
  cw.class = 0;
  cw.visual = 0;
  cw.mask = 0;
  writen(s, &cw, sizeof(cw));

  /* Make the window visible: */

  mw.req = 8; /* 8 means "show window" */
  mw.pad = 0;
  mw.length = sizeof(mw) >> 2;
  mw.id = id;
  write(s, &mw, sizeof(mw));

  while (1) { }
}
